##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import os
import unittest

import zope.configuration.xmlconfig
from zope.testing.cleanup import CleanUp

from z3c.pt import pagetemplate
from z3c.pt.pagetemplate import PageTemplateFile
from z3c.pt.pagetemplate import ViewPageTemplateFile


class Setup(CleanUp):
    def setUp(self):
        CleanUp.setUp(self)
        import z3c.pt

        zope.configuration.xmlconfig.file("configure.zcml", z3c.pt)


class TestPageTemplate(Setup, unittest.TestCase):
    def test_string_function(self):
        template = pagetemplate.PageTemplate(
            """<div tal:content="python:string('a string')" />"""
        )
        result = template.render()
        self.assertEqual(result, "<div>a string</div>")

    def test_nocall_function(self):
        class Call:
            def __call__(self):
                raise AssertionError("Should not be called")

            def __str__(self):
                return "Not Called"

        arg = {"call": Call()}
        template = pagetemplate.PageTemplate(
            """<div tal:content="python:nocall('arg/call')" />"""
        )
        result = template.render(arg=arg)
        self.assertEqual(result, "<div>Not Called</div>")

    def test_structure(self):
        template = pagetemplate.PageTemplate(
            "${structure: python: '&lt;div&gt;Hello world&lt;/div&gt;'}"
        )
        result = template.render()
        self.assertEqual(result, "<div>Hello world</div>")


class TestPageTemplateFile(Setup, unittest.TestCase):
    def test_nocall(self):
        template = PageTemplateFile("nocall.pt")

        def dont_call():
            raise AssertionError("Should not be called")

        result = template(callable=dont_call)
        self.assertIn(repr(dont_call), result)

    def test_exists(self):
        template = PageTemplateFile("exists.pt")

        def dont_call():
            raise AssertionError("Should not be called")

        result = template(callable=dont_call)
        self.assertIn("ok", result)

    def test_false_attribute(self):
        template = PageTemplateFile("false.pt")
        result = template()
        self.assertIn("False", result)

    def test_boolean_attribute(self):
        template = PageTemplateFile("boolean.pt")
        result = template()
        self.assertNotIn("False", result)
        self.assertIn('checked="checked"', result)

    def test_path(self):
        template = PageTemplateFile("path.pt")

        class Context:
            dummy_wysiwyg_support = True

        context = Context()
        template = template.__get__(context, Context)

        result = template(editor="dummy")
        self.assertIn("supported", result)
        self.assertIn("some path", result)


class TestViewPageTemplateFile(Setup, unittest.TestCase):
    def test_provider(self):
        class Context:
            pass

        class Request:
            response = None

        class View:
            __call__ = ViewPageTemplateFile("provider.pt")

        # Test binding descriptor behaviour.
        self.assertIsInstance(View.__call__, ViewPageTemplateFile)

        from zope.contentprovider.interfaces import ITALNamespaceData
        from zope.interface import Interface
        from zope.interface import directlyProvides
        from zope.interface import implementer
        from zope.schema import Field

        class ITestProvider(Interface):
            context = Field("Provider context.")

        directlyProvides(ITestProvider, ITALNamespaceData)
        assert ITALNamespaceData.providedBy(ITestProvider)

        @implementer(ITestProvider)
        class Provider:
            def __init__(self, *args):
                data.extend(list(args))

            def update(self):
                data.extend("updated")

            def render(self):
                return f"<![CDATA[ {data!r}, {self.__dict__!r}]]>"

        view = View()
        data = []

        from zope.component import provideAdapter
        from zope.contentprovider.interfaces import IContentProvider
        from zope.interface import implementedBy

        provideAdapter(
            Provider,
            (
                implementedBy(Context),
                implementedBy(Request),
                implementedBy(View),
            ),
            IContentProvider,
            name="content",
        )

        context = Context()
        request = Request()

        result = view(context=context, request=request)
        self.assertIn(repr(data), result)
        self.assertIn(repr({"context": context}), result)


class TestOpaqueDict(unittest.TestCase):
    def test_getitem(self):
        import operator

        d = {}
        od = pagetemplate.OpaqueDict(d)
        with self.assertRaises(KeyError):
            operator.itemgetter("key")(od)

        d["key"] = 42
        self.assertEqual(od["key"], 42)

    def test_len(self):
        d = {}
        od = pagetemplate.OpaqueDict(d)
        self.assertEqual(0, len(od))

        d["key"] = 42
        self.assertEqual(1, len(od))

    def test_repr(self):
        d = {}
        od = pagetemplate.OpaqueDict(d)
        self.assertEqual("{...} (0 entries)", repr(od))

        d["key"] = 42
        self.assertEqual("{...} (1 entries)", repr(od))


class TestBaseTemplate(unittest.TestCase):
    def test_negotiate_fails(self):
        class I18N:
            request = None

            def negotiate(self, request):
                self.request = request
                raise Exception("This is caught")

        i18n = I18N()
        orig_i18n = pagetemplate.i18n
        pagetemplate.i18n = i18n
        try:
            template = pagetemplate.BaseTemplate("<html />")
            request = "strings are allowed"
            template.render(request=request)
            self.assertIs(i18n.request, request)
        finally:
            pagetemplate.i18n = orig_i18n

    def test_translate_mv(self):
        template = pagetemplate.BaseTemplate(
            """
        <html>
          <body metal:use-macro="m" />
        </html>
        """
        )

        class Macro:
            translate = None

            def include(self, stream, econtext, *args, **kwargs):
                self.translate = econtext["translate"]

        macro = Macro()
        template.render(m=macro)

        self.assertIsNone(macro.translate(pagetemplate.MV))


class TestBaseTemplateFile(unittest.TestCase):
    def test_init_with_path(self):

        here = os.path.abspath(os.path.dirname(__file__))

        template = pagetemplate.BaseTemplateFile("view.pt", path=here)

        self.assertEqual(template.filename, os.path.join(here, "view.pt"))


class TestBoundPageTemplate(unittest.TestCase):

    def test_setattr(self):
        bound = pagetemplate.BoundPageTemplate(None, None)
        with self.assertRaisesRegex(AttributeError, "Can't set attribute"):
            bound.__self__ = 42

    def test_repr(self):
        # It requires the 'filename' attribute
        class Template:
            filename = "file.pt"

        bound = pagetemplate.BoundPageTemplate(Template(), "render")
        self.assertEqual(
            "<z3c.pt.tests.test_templates.BoundTemplate 'file.pt'>",
            repr(bound),
        )

    def test_attributes(self):
        func = object()
        bound = pagetemplate.BoundPageTemplate(self, func)
        self.assertIs(self, bound.im_self)
        self.assertIs(self, bound.__self__)
        self.assertIs(func, bound.im_func)
        self.assertIs(func, bound.__func__)



"""
Tests for expressions.py

"""
import functools
import unittest

from zope.testing.cleanup import CleanUp

from z3c.pt import expressions


# pylint:disable=protected-access


class TestRenderContentProvider(CleanUp, unittest.TestCase):
    def test_not_found(self):
        from zope.contentprovider.interfaces import ContentProviderLookupError

        context = object()
        request = object()
        view = object()
        name = "a provider"
        econtext = {"context": context, "request": request, "view": view}

        with self.assertRaises(ContentProviderLookupError) as exc:
            expressions.render_content_provider(econtext, name)

        e = exc.exception
        self.assertEqual(e.args, (name, (context, request, view)))

    def test_sets_ilocation_name(self):
        from zope import component
        from zope import interface
        from zope.contentprovider.interfaces import IContentProvider
        from zope.location.interfaces import ILocation

        attrs = {}

        @interface.implementer(ILocation, IContentProvider)
        class Provider:
            def __init__(self, *args):
                pass

            def __setattr__(self, name, value):
                attrs[name] = value

            update = render = lambda s: None

        component.provideAdapter(
            Provider,
            adapts=(object, object, object),
            provides=IContentProvider,
            name="a provider",
        )

        context = object()
        request = object()
        view = object()
        econtext = {"context": context, "request": request, "view": view}

        expressions.render_content_provider(econtext, "a provider")

        self.assertEqual(attrs, {"__name__": "a provider"})


class TestPathExpr(CleanUp, unittest.TestCase):
    def test_translate_empty_string(self):
        import ast

        expr = expressions.PathExpr("")
        result = expr.translate("", "foo")

        self.assertEqual(len(result), 1)
        self.assertIsInstance(result[0], ast.Assign)

    def test_translate_invalid_path(self):
        from chameleon.exc import ExpressionError

        expr = expressions.PathExpr("")
        with self.assertRaises(ExpressionError):
            expr.translate("not valid", None)

    def test_translate_components(self):
        from chameleon.compiler import ExpressionEngine
        from chameleon.compiler import ExpressionEvaluator
        from chameleon.tales import ExpressionParser
        from chameleon.utils import Scope

        parser = ExpressionParser({'path': expressions.PathExpr}, 'path')
        engine = functools.partial(ExpressionEngine, parser)

        from zope.traversing.interfaces import ITraversable

        class Context(str):
            def traverse(self, name, rest):
                return Context(name)

        from zope.interface import classImplements
        classImplements(Context, ITraversable)

        evaluator = ExpressionEvaluator(engine, {'a': Context('a')})

        def evaluate(expression, **context):
            return evaluator(Scope(context), {}, 'path', expression)

        self.assertEqual(evaluate("a"), Context('a'))

        # Single literal
        self.assertEqual(evaluate("a/b"), Context('b'))

        # Multiple literals
        self.assertEqual(evaluate("a/b/c"), Context('c'))

        # Single interpolation---must be longer than one character
        self.assertEqual(evaluate("a/?b"), Context('?b'))

        # Single interpolation
        self.assertEqual(evaluate("a/?t1", t1='b'), Context('b'))

        # Multiple interpolations
        self.assertEqual(evaluate("a/?t1/?t2", t1='b', t2='c'), Context('c'))



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import ast
import re
from types import MethodType

import zope.event
from chameleon.astutil import Builtin
from chameleon.astutil import NameLookupRewriteVisitor
from chameleon.astutil import Symbol
from chameleon.codegen import template
from chameleon.exc import ExpressionError
from chameleon.tales import ExistsExpr as BaseExistsExpr
from chameleon.tales import PythonExpr as BasePythonExpr
from chameleon.tales import StringExpr
from chameleon.tales import TalesExpr
from zope.contentprovider.interfaces import BeforeUpdateEvent
from zope.contentprovider.interfaces import ContentProviderLookupError
from zope.contentprovider.interfaces import IContentProvider
from zope.contentprovider.tales import addTALNamespaceData
from zope.location.interfaces import ILocation
from zope.traversing.adapters import traversePathElement
from zope.traversing.interfaces import ITraversable

import z3c.pt.namespaces


_marker = object()


def render_content_provider(econtext, name):
    name = name.strip()

    context = econtext.get("context")
    request = econtext.get("request")
    view = econtext.get("view")

    cp = zope.component.queryMultiAdapter(
        (context, request, view), IContentProvider, name=name
    )

    # provide a useful error message, if the provider was not found.
    # Be sure to provide the objects in addition to the name so
    # debugging ZCML registrations is possible
    if cp is None:
        raise ContentProviderLookupError(name, (context, request, view))

    # add the __name__ attribute if it implements ILocation
    if ILocation.providedBy(cp):
        cp.__name__ = name

    # Insert the data gotten from the context
    addTALNamespaceData(cp, econtext)

    # Stage 1: Do the state update.
    zope.event.notify(BeforeUpdateEvent(cp, request))
    cp.update()

    # Stage 2: Render the HTML content.
    return cp.render()


def path_traverse(base, econtext, call, path_items):
    if path_items:
        request = econtext.get("request")
        path_items = list(path_items)
        path_items.reverse()

        while path_items:
            name = path_items.pop()
            ns_used = ":" in name
            if ns_used:
                namespace, name = name.split(":", 1)
                base = z3c.pt.namespaces.function_namespaces[namespace](base)
                if ITraversable.providedBy(base):
                    base = traversePathElement(
                        base, name, path_items, request=request
                    )

                    # base = proxify(base)

                    continue

            # special-case dicts for performance reasons
            if isinstance(base, dict):
                next = base.get(name, _marker)
            else:
                next = getattr(base, name, _marker)

            if next is not _marker:
                base = next
                if ns_used and isinstance(base, MethodType):
                    base = base()
                # The bytecode peephole optimizer removes the next line:
                continue  # pragma: no cover
            else:
                base = traversePathElement(
                    base, name, path_items, request=request
                )

            # if not isinstance(base, (basestring, tuple, list)):
            #    base = proxify(base)

    if call and getattr(base, "__call__", _marker) is not _marker:
        return base()

    return base


class ContextExpressionMixin:
    """Mixin-class for expression compilers."""

    transform = None

    def __call__(self, target, engine):
        # Make call to superclass to assign value to target
        assignment = super().__call__(
            target, engine
        )

        transform = template(
            "target = transform(econtext, target)",
            target=target,
            transform=self.transform,
        )

        return assignment + transform


class PathExpr(TalesExpr):
    path_regex = re.compile(
        r"^(?:(nocall|not):\s*)*((?:[A-Za-z0-9_][A-Za-z0-9_:]*)"
        + r"(?:/[?A-Za-z0-9_@\-+][?A-Za-z0-9_@\-\.+/:]*)*)$"
    )

    interpolation_regex = re.compile(r"\?[A-Za-z][A-Za-z0-9_]+")

    traverser = Symbol(path_traverse)

    def _find_components(self, parts):
        components = []
        for part in parts[1:]:
            interpolation_args = []

            def replace(match):
                start, end = match.span()
                interpolation_args.append(part[start + 1: end])
                return "%s"

            while True:
                part, count = self.interpolation_regex.subn(replace, part)
                if count == 0:
                    break

            if interpolation_args:
                component = template(
                    "format % args",
                    format=ast.Constant(part),
                    args=ast.Tuple(
                        [ast.Name(arg, ctx=ast.Load())
                         for arg in interpolation_args],
                        ast.Load(),
                    ),
                    mode="eval",
                )
            else:
                component = ast.Constant(part)

            components.append(component)

        return components

    def translate(self, string, target):
        """
        >>> from chameleon.tales import test
        >>> test(PathExpr('None')) is None
        True
        """
        string = string.strip()

        if not string:
            return template("target = None", target=target)

        m = self.path_regex.match(string)
        if m is None:
            raise ExpressionError("Not a valid path-expression.", string)

        nocall, path = m.groups()

        # note that unicode paths are not allowed
        parts = str(path).split("/")

        components = self._find_components(parts)

        base = parts[0]

        if not components:
            if len(parts) == 1 and (nocall or base == "None"):
                return template("target = base", base=base, target=target)
            else:
                components = ()

        call = template(
            "traverse(base, econtext, call, path_items)",
            traverse=self.traverser,
            base=base,
            call=str(not nocall),
            path_items=ast.Tuple(elts=components),
            mode="eval",
        )

        return template("target = value", target=target, value=call)


class NocallExpr(PathExpr):
    """A path-expression which does not call the resolved object."""

    def translate(self, expression, engine):
        return super().translate(
            "nocall:%s" % expression, engine
        )


class ExistsExpr(BaseExistsExpr):
    exceptions = AttributeError, LookupError, TypeError, KeyError, NameError

    def __init__(self, expression):
        super().__init__("nocall:" + expression)


class ProviderExpr(ContextExpressionMixin, StringExpr):
    transform = Symbol(render_content_provider)


class PythonExpr(BasePythonExpr):
    builtins = {
        name: template(
            "tales(econtext, rcontext, name)",
            tales=Builtin("tales"),
            name=ast.Constant(name),
            mode="eval",
        )
        for name in ("path", "exists", "string", "nocall")
    }

    def __call__(self, target, engine):
        return self.translate(self.expression, target)

    def rewrite(self, node):
        builtin = self.builtins.get(node.id)
        if builtin is not None:
            return template(
                "get(name) if get(name) is not None else builtin",
                get=Builtin("get"),
                name=ast.Constant(node.id),
                builtin=builtin,
                mode="eval",
            )

    @property
    def transform(self):
        return NameLookupRewriteVisitor(self.rewrite)



##############################################################################
#
# Copyright (c) 2023 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest

import zope.component.testing
import zope.configuration.xmlconfig
from chameleon.tal import ErrorInfo
from chameleon.tal import RepeatItem

import z3c.pt
from z3c.pt.interfaces import ITALESIterator
from z3c.pt.interfaces import ITALExpressionErrorInfo


def setUp(suite):
    zope.component.testing.setUp(suite)
    zope.configuration.xmlconfig.XMLConfig("configure.zcml", z3c.pt)()


class TestRenderContentProvider(unittest.TestCase):
    setUp = setUp
    tearDown = zope.component.testing.tearDown

    def test_implements(self):
        self.assertTrue(ITALExpressionErrorInfo.implementedBy(ErrorInfo))
        self.assertTrue(ITALESIterator.implementedBy(RepeatItem))



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import os
import sys

from chameleon.compiler import ExpressionEvaluator
from chameleon.i18n import fast_translate
from chameleon.tales import NotExpr
from chameleon.tales import StringExpr
from chameleon.tales import StructureExpr
from chameleon.zpt import template
from zope import i18n
from zope.security.proxy import ProxyFactory

from z3c.pt import expressions


try:
    from Missing import MV

    MV = MV  # pragma: no cover
except ModuleNotFoundError:
    MV = object()

_marker = object()


BOOLEAN_HTML_ATTRS = frozenset(
    [
        # List of Boolean attributes in HTML that should be rendered in
        # full form, e.g., <img ismap="ismap"> rather than their minimized
        # form <img ismap>, as required by XML 1.0.
        # From https://www.w3.org/TR/xhtml1/#C_10 and
        # https://www.w3.org/TR/xhtml1/#h-4.5
        # Updated to also include html5 attributes as listed on
        # https://html.spec.whatwg.org/multipage/indices.html#attributes-3
        "allowfullscreen",
        "async",
        "autofocus",
        "autoplay",
        "checked",
        "compact",
        "controls",
        "declare",
        "default",
        "defer",
        "disabled",
        "formnovalidate",
        "inert",
        "ismap",
        "itemscope",
        "loop",
        "multiple",
        "muted",
        "nomodule",
        "noresize",
        "noshade",
        "novalidate",
        "nowrap",
        "open",
        "playsinline",
        "readonly",
        "required",
        "reversed",
        "selected",
        "shadowrootclonable",
        "shadowrootdelegatesfocus",
        "shadowrootserializable",
    ]
)


class OpaqueDict(dict):
    def __new__(cls, dictionary):
        inst = dict.__new__(cls)
        inst.dictionary = dictionary
        return inst

    def __getitem__(self, name):
        return self.dictionary[name]

    def __len__(self):
        return len(self.dictionary)

    def __repr__(self):
        return "{...} (%d entries)" % len(self)


sys_modules = ProxyFactory(OpaqueDict(sys.modules))


class BaseTemplate(template.PageTemplate):
    content_type = None
    version = 2

    expression_types = {
        "python": expressions.PythonExpr,
        "string": StringExpr,
        "not": NotExpr,
        "exists": expressions.ExistsExpr,
        "path": expressions.PathExpr,
        "provider": expressions.ProviderExpr,
        "nocall": expressions.NocallExpr,
        "structure": StructureExpr,
    }

    default_expression = "path"

    literal_false = True

    strict = False

    trim_attribute_space = True

    # https://github.com/zopefoundation/Zope/issues/716
    enable_comment_interpolation = False

    @property
    def boolean_attributes(self):
        if self.content_type == "text/xml":
            return set()

        return BOOLEAN_HTML_ATTRS

    @property
    def builtins(self):
        builtins = {"nothing": None, "modules": sys_modules}

        tales = ExpressionEvaluator(self.engine, builtins)
        builtins["tales"] = tales

        return builtins

    def bind(self, ob, request=None):
        def render(request=request, **kwargs):
            context = self._pt_get_context(ob, request, kwargs)
            return self.render(**context)

        return BoundPageTemplate(self, render)

    def render(self, target_language=None, **context):
        # We always include a ``request`` variable; it is (currently)
        # depended on in various expression types and must be defined
        request = context.setdefault("request", None)

        if target_language is None:
            try:
                target_language = i18n.negotiate(request)
            except Exception:
                target_language = None

        context["target_language"] = target_language

        # bind translation-method to request
        def translate(
            msgid,
            domain=None,
            mapping=None,
            target_language=None,
            default=None,
            context=None,
        ):
            if msgid is MV:
                # Special case handling of Zope2's Missing.MV
                # (Missing.Value) used by the ZCatalog but is
                # unhashable.

                # This case cannot arise in ordinary templates; msgid
                # comes from i18n:translate attributes, which does not
                # take a TALES expression, just a literal string.
                # However, the 'context' argument is available as an
                # implementation detail for macros
                return
            return fast_translate(
                msgid, domain, mapping, request, target_language, default
            )

        context["translate"] = translate

        if request is not None and not isinstance(request, str):
            content_type = self.content_type or "text/html"
            response = request.response
            if response and not response.getHeader("Content-Type"):
                response.setHeader("Content-Type", content_type)

        base_renderer = super().render
        return base_renderer(**context)

    def __call__(self, *args, **kwargs):
        bound_pt = self.bind(self)
        return bound_pt(*args, **kwargs)

    def _pt_get_context(self, instance, request, kwargs):
        return dict(
            context=instance,
            here=instance,
            options=kwargs,
            request=request,
            template=self,
        )


class BaseTemplateFile(BaseTemplate, template.PageTemplateFile):
    """If ``filename`` is a relative path, the module path of the
    class where the instance is used to get an absolute path."""

    cache = {}

    def __init__(self, filename, path=None, content_type=None, **kwargs):
        if path is not None:
            filename = os.path.join(path, filename)

        if not os.path.isabs(filename):
            for depth in (1, 2):
                frame = sys._getframe(depth)
                package_name = frame.f_globals.get("__name__", None)
                if (
                    package_name is not None
                    and package_name != self.__module__
                ):
                    module = sys.modules[package_name]
                    try:
                        path = module.__path__[0]
                    except AttributeError:
                        path = module.__file__
                        path = path[: path.rfind(os.sep)]
                    break
                else:
                    package_path = frame.f_globals.get("__file__", None)
                    if package_path is not None:
                        path = os.path.dirname(package_path)
                        break

            if path is not None:
                filename = os.path.join(path, filename)

        template.PageTemplateFile.__init__(self, filename, **kwargs)

        # Set content-type last, so that we can override whatever was
        # magically sniffed from the source template.
        self.content_type = content_type


class PageTemplate(BaseTemplate):
    """Page Templates using TAL, TALES, and METAL.

    This class is suitable for standalone use or class
    property. Keyword-arguments are passed into the template as-is.

    Initialize with a template string."""

    version = 1

    def __get__(self, instance, type):
        if instance is not None:
            return self.bind(instance)
        return self


class PageTemplateFile(BaseTemplateFile, PageTemplate):
    """Page Templates using TAL, TALES, and METAL.

    This class is suitable for standalone use or class
    property. Keyword-arguments are passed into the template as-is.

    Initialize with a filename."""

    cache = {}


class ViewPageTemplate(PageTemplate):
    """Template class suitable for use with a Zope browser view; the
    variables ``view``, ``context`` and ``request`` variables are
    brought in to the local scope of the template automatically, while
    keyword arguments are passed in through the ``options``
    dictionary. Note that the default expression type for this class
    is 'path' (standard Zope traversal)."""

    def _pt_get_context(self, view, request, kwargs):
        context = kwargs.get("context")
        if context is None:
            context = view.context
        request = request or kwargs.get("request") or view.request
        return dict(
            view=view,
            context=context,
            request=request,
            options=kwargs,
            template=self,
        )

    def __call__(self, _ob=None, context=None, request=None, **kwargs):
        kwargs.setdefault("context", context)
        kwargs.setdefault("request", request)
        bound_pt = self.bind(_ob)
        return bound_pt(**kwargs)


class ViewPageTemplateFile(ViewPageTemplate, PageTemplateFile):
    """If ``filename`` is a relative path, the module path of the
    class where the instance is used to get an absolute path."""

    cache = {}


class BoundPageTemplate:
    """When a page template class is used as a property, it's bound to
    the class instance on access, which is implemented using this
    helper class."""

    __self__ = None
    __func__ = None

    def __init__(self, pt, render):
        object.__setattr__(self, "__self__", pt)
        object.__setattr__(self, "__func__", render)

    im_self = property(lambda self: self.__self__)
    im_func = property(lambda self: self.__func__)
    macros = property(lambda self: self.__self__.macros)
    filename = property(lambda self: self.__self__.filename)

    def __call__(self, *args, **kw):
        kw.setdefault("args", args)
        return self.__func__(**kw)

    def __setattr__(self, name, v):
        raise AttributeError("Can't set attribute", name)

    def __repr__(self):
        return "<{}.Bound{} {!r}>".format(
            type(self.__self__).__module__,
            type(self.__self__).__name__,
            self.filename,
        )



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import doctest
import unittest

import zope.component.testing
import zope.configuration.xmlconfig

import z3c.pt


OPTIONFLAGS = doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE


def setUp(suite):
    zope.component.testing.setUp(suite)
    zope.configuration.xmlconfig.XMLConfig("configure.zcml", z3c.pt)()


def test_suite():
    return unittest.TestSuite(
        [
            doctest.DocFileSuite(
                "README.rst",
                optionflags=OPTIONFLAGS,
                setUp=setUp,
                tearDown=zope.component.testing.tearDown,
                package="z3c.pt",
            ),
            doctest.DocTestSuite(
                "z3c.pt.expressions",
                optionflags=OPTIONFLAGS,
                setUp=setUp,
                tearDown=zope.component.testing.tearDown,
            ),
            doctest.DocTestSuite(
                "z3c.pt.namespaces",
                optionflags=OPTIONFLAGS,
                setUp=setUp,
                tearDown=zope.component.testing.tearDown,
            ),
        ]
    )



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest


class LoadTests:
    def _makeOne(self, search_path=None, **kwargs):
        klass = self._getTargetClass()
        return klass(search_path, **kwargs)

    def _getTargetClass(self):
        from z3c.pt.loader import TemplateLoader

        return TemplateLoader

    def test_load_relative(self):
        import os

        here = os.path.dirname(__file__)
        loader = self._makeOne(search_path=[here])

        result = self._load(loader, "view.pt")
        self.assertEqual(result.filename, os.path.join(here, "view.pt"))

    def test_consecutive_loads(self):
        import os

        here = os.path.dirname(__file__)
        loader = self._makeOne(search_path=[here])

        self.assertIs(
            self._load(loader, "view.pt"),
            self._load(loader, "view.pt")
        )

    def test_load_relative_badpath_in_searchpath(self):
        import os

        here = os.path.dirname(__file__)
        loader = self._makeOne(search_path=[os.path.join(here, "none"), here])
        result = self._load(loader, "view.pt")
        self.assertEqual(result.filename, os.path.join(here, "view.pt"))

    def test_load_abs(self):
        import os

        here = os.path.dirname(__file__)
        loader = self._makeOne()
        abs = os.path.join(here, "view.pt")
        result = self._load(loader, abs)
        self.assertEqual(result.filename, abs)


class LoadPageTests(unittest.TestCase, LoadTests):
    def _load(self, loader, filename):
        return loader.load_page(filename)



"""
Tests for namespaces.py.

"""

import unittest

from z3c.pt import namespaces


class TestAdapterNamespaces(unittest.TestCase):
    def test_registered_name(self):
        nss = namespaces.AdapterNamespaces()

        def func(ctx):
            return ctx
        nss.registerFunctionNamespace("ns", func)
        self.assertIs(func, nss["ns"])
        self.assertIs(func, nss.getFunctionNamespace("ns"))

    def test_unregistered_name(self):
        nss = namespaces.AdapterNamespaces()
        with self.assertRaises(KeyError):
            nss.getFunctionNamespace("ns")

        # but __getitem__ makes one up
        self.assertIsNotNone(nss["ns"])

    def test_using_pagetemplate_version(self):
        from zope.pagetemplate import engine

        self.assertNotIsInstance(
            namespaces.function_namespaces, namespaces.AdapterNamespaces
        )
        self.assertIsInstance(
            namespaces.function_namespaces, engine.AdapterNamespaces
        )



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Setup
"""
import os

from setuptools import setup


def read(*filenames):
    with open(os.path.join(*filenames)) as f:
        return f.read()


setup(
    name="z3c.pt",
    version='5.0',
    author="Malthe Borch and the Zope Community",
    author_email="zope-dev@zope.dev",
    description="Fast ZPT engine.",
    long_description=(
        read("README.rst")
        + "\n\n"
        + read("src", "z3c", "pt", "README.rst")
        + "\n\n"
        + read("CHANGES.rst")
    ),
    license="ZPL",
    keywords="tal tales pagetemplate zope chameleon",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Zope Public License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Text Processing :: Markup :: HTML",
        "Topic :: Text Processing :: Markup :: XML",
        "Framework :: Zope :: 3",
    ],
    url="https://github.com/zopefoundation/z3c.pt",
    project_urls={
        'Sources': 'https://github.com/zopefoundation/z3c.pt',
        'Issue Tracker': 'https://github.com/zopefoundation/z3c.pt/issues',
    },
    python_requires='>=3.9',
    install_requires=[
        "setuptools",
        "zope.interface",
        "zope.component",
        "zope.i18n >= 3.5",
        "zope.traversing",
        "zope.contentprovider",
        "Chameleon >= 2.4",
    ],
    extras_require={
        "test": [
            "zope.pagetemplate",
            "zope.testing",
            "zope.testrunner >= 6.4",
        ],
        "docs": [
            "Sphinx",
        ]
    },
    include_package_data=True,
    zip_safe=False,
)
